#!/usr/bin/env python
#
# frep_gpu.py
#    functional representation solver - OpenGL version
#
# usage:
#    pcb.py | frep.py [dpi [filename]]
#
# Neil Gershenfeld 9/30/19
# (c) Massachusetts Institute of Technology 2019
# Quentin Bolsée 25/04/20
#
# This work may be reproduced, modified, distributed,
# performed, and displayed for any purpose, but must
# acknowledge this project. Copyright is retained and
# must be preserved. The work is provided as is; no
# warranty is provided, and users accept all liability.
#

from OpenGL.GLUT import *
from OpenGL.GL import *
import OpenGL.GL.shaders as shaders
from PIL import Image
import json
import numpy as np

txt_vert = """#version 420

uniform vec2 x_lims;
uniform vec2 y_lims;
uniform float z;

layout (location = 0) in vec3 pos;
out vec3 v_pos;

float map(float value, float min1, float max1, float min2, float max2) {
    return min2 + (value - min1) * (max2 - min2) / (max1 - min1);
}

void main() {
    gl_Position = vec4(pos, 1.0);
    float x, y;

    x = map(pos.x, -1, 1, x_lims.s, x_lims.t);
    y = map(pos.y, -1, 1, y_lims.s, y_lims.t);
    v_pos = vec3(x, y, z);
}
"""

txt_frag = """#version 420
uniform vec3 color;
uniform float alpha;

in vec3 v_pos;
out vec4 f_color;

void main() {
    float X, Y, Z;
    X = v_pos.x;
    Y = v_pos.y;
    Z = v_pos.z;

    bool c = [F_LOGIC];

    f_color = vec4(color, float(c)*alpha);
}
"""


def generate_program(f):
    global txt_vert, txt_frag
    shader_vert = shaders.compileShader(txt_vert, GL_VERTEX_SHADER)
    shader_frag = shaders.compileShader(txt_frag.replace("[F_LOGIC]", f), GL_FRAGMENT_SHADER)
    return shaders.compileProgram(shader_vert, shader_frag)


class frepRender:
    def __init__(self, width, height, dpi, filename):
        self.width = width
        self.height = height
        self.init_window()
        self.vao = 0
        self.vbo = 0
        self.vbo_data = None
        self.fbo = 0
        self.rbo_rgb = 0
        self.init_buffer()
        self.x_lims = (0, 1)
        self.y_lims = (0, 1)
        self.z_list = []
        self.functions_list = []
        self.shader_list = []
        self.init_vbo()
        self.filename = filename
        self.dpi = dpi
        self.n_frames = 1

    def init_window(self):
        glutInit()
        glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH)
        glutInitWindowSize(self.width, self.height)
        # glutInitWindowPosition(100, 100)
        glutCreateWindow("frep OpenGL")

    def init_buffer(self):
        self.fbo = glGenFramebuffers(1)
        self.rbo_rgb = glGenRenderbuffers(1)

        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)
        glBindRenderbuffer(GL_RENDERBUFFER, self.rbo_rgb)
        glRenderbufferStorage(GL_RENDERBUFFER, GL_RGBA8, self.width, self.height)
        glBindRenderbuffer(GL_RENDERBUFFER, 0)

        glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
                                  GL_RENDERBUFFER, self.rbo_rgb)
        status = glCheckFramebufferStatus(GL_FRAMEBUFFER)
        assert status == GL_FRAMEBUFFER_COMPLETE
        glBindFramebuffer(GL_FRAMEBUFFER, 0)

    def set_functions(self, f_list):
        self.functions_list = f_list
        for f in f_list:
            prog = generate_program(f["f"])
            self.shader_list.append(prog)

    def set_geometry(self, x_lims, y_lims, z_list):
        self.x_lims = x_lims
        self.y_lims = y_lims
        self.z_list = z_list

    def init_vbo(self):
        self.vao = glGenVertexArrays(1)
        glBindVertexArray(self.vao)
        # a single triangle strip covering the screen
        vbo_data = np.array([-1, -1, 0,
                             -1, 1, 0,
                             1, -1, 0,
                             1, 1, 0], dtype=np.float32)
        self.vbo_data = vbo_data
        self.vbo = glGenBuffers(1)
        glBindBuffer(GL_ARRAY_BUFFER, self.vbo)
        glBufferData(GL_ARRAY_BUFFER, vbo_data.nbytes, vbo_data, GL_STATIC_DRAW)

        glEnableVertexAttribArray(0)
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3*vbo_data.itemsize, ctypes.c_void_p(0))

        glBindVertexArray(0)
        glBindBuffer(GL_ARRAY_BUFFER, 0)

    def display(self):
        w, h = self.width, self.height

        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)
        # glEnable(GL_DEPTH_TEST)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glClearColor(0.0, 0.0, 0.0, 1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glBindVertexArray(self.vao)
        for z in self.z_list:
            for i in range(len(self.shader_list)):
                prog = self.shader_list[i]
                f = self.functions_list[i]
                if not f["active"]:
                    continue
                color = f["color"]
                rgb = (((color & 0xFF0000) >> 16)/255.0, ((color & 0x00FF00) >> 8)/255.0, (color & 0x0000FF)/255.0)
                alpha = f["alpha"]

                glUseProgram(prog)
                loc_color = glGetUniformLocation(prog, "color")
                loc_alpha = glGetUniformLocation(prog, "alpha")
                loc_x_lims = glGetUniformLocation(prog, "x_lims")
                loc_y_lims = glGetUniformLocation(prog, "y_lims")
                loc_z = glGetUniformLocation(prog, "z")
                glUniform3f(loc_color, *rgb)
                glUniform1f(loc_alpha, alpha)
                glUniform2f(loc_x_lims, *self.x_lims)
                glUniform2f(loc_y_lims, *self.y_lims)
                glUniform1f(loc_z, z)

                glDrawArrays(GL_TRIANGLE_STRIP, 0, 4)
        glBindVertexArray(0)

        glBindFramebuffer(GL_FRAMEBUFFER, 0)

        glBindFramebuffer(GL_READ_FRAMEBUFFER, self.fbo)
        glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0)
        glBlitFramebuffer(0, 0, w, h, 0, 0, w, h, GL_COLOR_BUFFER_BIT, GL_NEAREST)
        glutSwapBuffers()

        self.n_frames -= 1
        if self.n_frames == 0:
            self.stop()

    def get_render(self):
        glBindFramebuffer(GL_FRAMEBUFFER, self.fbo)
        rgb_bytes = glReadPixels(0, 0, self.width, self.height, GL_RGB, GL_UNSIGNED_BYTE)
        glBindFramebuffer(GL_FRAMEBUFFER, 0)
        rgb = np.frombuffer(rgb_bytes, dtype=np.uint8).reshape(self.height, self.width, 3)
        return np.flip(rgb, 0)

    def launch(self, n_frames):
        self.n_frames = n_frames
        glutDisplayFunc(self.display)
        glutKeyboardFunc(self.keyboard)
        glutIdleFunc(self.update)
        glutMainLoop()

    def update(self):
        glutPostRedisplay()

    def keyboard(self, key, x, y):
        key = str(key, "ascii")
        if key == chr(27):
            self.stop()

    def stop(self):
        if self.filename != "":
            rgb = self.get_render()
            im = Image.fromarray(rgb, 'RGB')
            im.save(self.filename, dpi=[self.dpi, self.dpi])
        sys.exit(0)


def main():
    frep = json.load(sys.stdin)

    #
    # check arguments
    #
    if (frep['type'] != 'RGB'):
        print('types other than RGB not (yet) supported')
        sys.exit()
    if (len(sys.argv) == 1):
        print('output to out.png at 100 DPI')
        filename = 'out.png'
        dpi = 100
    elif (len(sys.argv) == 2):
        dpi = sys.argv[1]
        filename = 'out.png'
        print('output to out.png at '+dpi+'DPI')
        dpi = int(dpi)
    elif (len(sys.argv) == 3):
        dpi = sys.argv[1]
        filename = sys.argv[2]
        print('output to '+filename+' at '+dpi+' DPI')
        dpi = int(dpi)
    else:
        dpi = 100
        filename = ""

    #
    # evaluate
    #
    print('evaluating')
    xmin = frep['xmin']
    xmax = frep['xmax']
    ymin = frep['ymin']
    ymax = frep['ymax']
    z_list = frep['layers']
    units = float(frep['mm_per_unit'])
    delta = (25.4/dpi)/units

    width = round((xmax-xmin)/delta)
    height = round((ymax-ymin)/delta)

    window = frepRender(width, height, dpi, filename)
    window.set_geometry((xmin, xmax),
                        (ymin, ymax),
                        z_list)
    window.set_functions(frep['functions'])

    # launch for only 1 frame, then save the file
    window.launch(1)


if __name__ == "__main__":
    main()
